%% SPINTA E ANG VARIABILE NEL TRATTO DISCESA . ANG VARIABILE ATTERRAGGIO 100+100+10+2
% Secondo tratto ottimizzata solo la spinta con angolo fisso a 90
clc;
clear all;

global G M_moon R_moon r_rif v_rif m_rif t_rif a_rif;
G = 6.67e-11;   % m^3/kg/s^2
M_moon = 7.342e22; % kg
R_moon = 1.738e6;  % m
r_rif = 15000;     % m
v_rif = 1000;       % m/s
m_rif = 1600;      % kg
t_rif = r_rif / v_rif; % s
a_rif = v_rif^2./ r_rif;

% Condizioni iniziali
y0 = [-500; 0; 15000; sqrt(G*M_moon/(R_moon + r_rif)); 0; 0; 1600]; % Dimensionale
y0_adim = y0 ./ [r_rif; r_rif; r_rif; v_rif; v_rif; v_rif; m_rif]; % Adimensionale
problems.x0 = y0_adim; % Stato iniziale del sistema

% Inizializzazione dei vincoli sul tempo finale
problems.tfMin = 30;  % Limite inferiore per tf_adim
problems.tfMax = 40;% Limite superiore per tf_adim
problems.tfMin1 = 36;  % Limite inferiore per tf_adim
problems.tfMax1 = 41;% Limite superiore per tf_adim

% Impostazione dei vincoli sui controlli 
problems.controlMin = [2]; % Limite inferiore per il controllo
problems.controlMax = [pi];  % Limite superiore per il controllo
problems.controlMin1 = [1.4]; % Limite inferiore per il controllo
problems.controlMax1 = [1.7];  % Limite superiore per il controllo
problems.uguess(:, 1) = [2.4; 2.95]; % Stima iniziale per i controlli
problems.u1guess(:, 1) = [1.65; 1.56];
problems.tguess = [0; 30]; % Stima iniziale per il tempo
problems.t1guess = [0; 40]; % Stima iniziale per il tempo
problems.nodes = 100; % Numero di nodi per la discretizzazione temporale
problems.nodes1 = 10;
problems.nodestot = 2.*problems.nodes+problems.nodes;

problems.spintaMin = [3600./(m_rif*a_rif)]; % Limite inferiore per il controllo sulla spinta
problems.spintaMax = [4000./(m_rif*a_rif)];  % Limite superiore per il controllo
problems.spintaguess(:,1) = [3900./(m_rif*a_rif); 3700/(m_rif*a_rif)];

problems.tguess = [0; 30]; % Stima iniziale per il tempo
problems.t1guess = [0; 40]; % Stima iniziale per il tempo

solution = risolvi(problems);
fprintf('La massa finale è: %.2f kg\n', solution.state(end, 7) * m_rif);
fprintf('Il tempo di discesa è: %.2f s\n', solution.tf_tf1(1)*t_rif)
fprintf('Il tempo di touchdown è: %.2f s\n', solution.tf_tf1(2)*t_rif)

figure(1); 
clf;
hold on
plot(solution.time, solution.state(:, 6), 'o'); % Stato Vz (punti)
% hold on
% plot(solution.time, solution.state(:, 4), 'o'); % Stato Vx (punti)
% xline(solution.tf_tf1(1),'--r',{'Tempo discesa','tf'});
% xline(solution.tf_tf1(2),'--r',{'Tempo approccio','tf1'});
hold off;
grid on; % Attiva la griglia
xlabel('Tempo Adimensionale [/]'); % Etichetta per l'asse x
ylabel('Velocità verticale [/]'); % Etichetta per l'asse y
title('Velocità verticale')

figure(2);
plot(solution.state(:, 1), solution.state(:, 3), 'o'); % Traiettoria discreta
grid on; % Attiva la griglia
xlabel('Stato x [/]'); % Etichetta per l'asse x
ylabel('Altitudine z [/]'); % Etichetta per l'asse y
title('Discesa da h_0=15km')

figure(3);
plot(solution.state(:, 1), solution.state(:, 3), 'o'); % Traiettoria discreta
yline(20./r_rif,'--r',{'Approccio verticale','20m/r_rif'});
xlim([solution.state(end, 1)-0.01 solution.state(end, 1)+0.01]);
ylim([0 0.005]);
grid on; % Attiva la griglia
xlabel('Stato x [/]'); % Etichetta per l'asse x
ylabel('Altitudine z [/]'); % Etichetta per l'asse y
title('Atterraggio di 5s')

figure(4); clf;
plot(solution.time(1:110), solution.control(1:110), 'o'); % Controllo discreto
grid on; % Attiva la griglia
yline(1.5708,'--r',{'Verticale','90°'});
xlabel('Tempo Adimensionale [/]'); % Etichetta per l'asse x
ylabel('Angolo di spinta [rad]'); % Etichetta per l'asse y
title('Angolo di controllo');

figure(5); clf;
plot(solution.time(1:100), solution.control(111:210).*(m_rif*a_rif), 'o'); % Controllo discreto
grid on; % Attiva la griglia
xlabel('Tempo Adimensionale [/]'); % Etichetta per l'asse x
ylabel('Spinta [N]'); % Etichetta per l'asse y
title('Spinta ottimale');

figure(6); clf;
hold on
plot(solution.time(101:110).*t_rif, solution.state(101:110, 6).*v_rif, 'o'); % Stato x (punti)
hold off;
grid on; % Attiva la griglia
xlabel('Tempo [s]'); % Etichetta per l'asse x
ylabel('Velocità verticale [m/s]'); % Etichetta per l'asse y
title('Velocità verticale di atterraggio di 5s @ T_d_i_s_c variabile')
function solutionOut=risolvi(problem)
global t_rif m_rif a_rif
    problem.ns = length(problem.x0);
    problem.nc = length(problem.controlMin);
    problem.nuVar = problem.nc * (2.*problem.nodes+problem.nodes1); % Variabili di controllo totali (210 angoli piu 10 spinte)
    problem.npts = linspace(0, 1, problem.nodes); % Punti normalizzati
    problem.npts1 = linspace(0, 1, problem.nodes1); % Punti normalizzati

    % 4. Indici delle variabili
    problem.numVar = problem.nuVar + 2;  % Numero totale di variabili (controlli 210 + t1 + tf)
    problem.nuIndex = 1:110;   % Indici per le angoli di controllo
    problem.nuTIndex = 111:210;
    problem.ntIndex = problem.nuVar + 1; % Indice per tf
    problem.ntIndex1 = problem.nuVar + 2;% Indice per tf

    % 5. Interpolazione delle variabili di controllo
    uVar = interp1(problem.tguess / problem.tguess(end), problem.uguess, problem.npts, "linear");
    uVar1 = interp1(problem.t1guess / problem.t1guess(end), problem.u1guess, problem.npts1, "linear");
    spintaVar = interp1(problem.t1guess / problem.t1guess(end), problem.spintaguess, problem.npts, "spline");

    % 6. Inizializzazione delle variabili
    allVarGuess = [uVar(:); uVar1(:); spintaVar(:); problem.tguess(end); problem.t1guess(end)]; % Insieme delle variabili iniziali
    varMin = repmat(problem.controlMin, 1, problem.nodes)'; % Limiti inferiori
    varMax = repmat(problem.controlMax, 1, problem.nodes)'; % Limiti superiori
    varMin1 = repmat(problem.controlMin1, 1, problem.nodes1)'; % Limiti inferiori
    varMax1 = repmat(problem.controlMax1, 1, problem.nodes1)'; % Limiti superiori
    varSMin1 = repmat(problem.spintaMin, 1, problem.nodes)'; % Limiti inferiori
    varSMax1 = repmat(problem.spintaMax, 1, problem.nodes)'; % Limiti superiori
    allVarLower = [varMin(:);  varMin1(:); varSMin1(:); problem.tfMin; problem.tfMin1]; % Limiti inferiori totali
    allVarUpper = [varMax(:); varMax1(:); varSMax1(:); problem.tfMax; problem.tfMax1]; % Limiti superiori totali

    optODE = odeset('RelTol', 1e-3, 'AbsTol', 1e-3); % Opzioni per l'ODE
    problem.optODE = optODE;
    
    % 9. Opzioni per l'ottimizzazione
    options = optimoptions('fmincon', ...
            'Algorithm', 'sqp', ...                    % Cambiamo l'algoritmo a 'sqp'
            'ConstraintTolerance', 1e-6, ...           % Rilassiamo la tolleranza sui vincoli
            'StepTolerance', 1e-6, ...                 % Aumentiamo la tolleranza sui passi
            'Display', 'iter-detailed', ...            % Mostra dettagli sull'ottimizzazione ad ogni iterazione
            'MaxIterations', 2000, ...                 % Aumentiamo il numero massimo di iterazioni
            'OptimalityTolerance', 1e-4, ...           % Tolleranza per la condizione di ottimalità
            'FunctionTolerance', 1e-4);                % Tolleranza per il valore della funzione obiettivo
    
    % 10. Risoluzione del problema di ottimizzazione
    t1 = tic;          % Inizio cronometro
    x = fmincon(@(allVar) objectivee(allVar, problem), allVarGuess, [], [], [], [], allVarLower, allVarUpper, @(allVar) constraintss(allVar, problem), options);
    solTime = toc(t1); % Tempo di soluzione
    
    %11. Estrazione dei risultati
    controlRet = x(1:110); %1=1:110
    TRet= x(111:210); %1:100
    tf = x(211);
    tf1= x(212);
    t_combined = [];
    y_combined = [];
    
    tspan=tf/problem.nodes;
    y_prev=problem.x0;
    nod_tot=problem.nodes+problem.nodes1;
    %T=4000.*ones(1,problem.nodes);
    psi=deg2rad(zeros(1,nod_tot)); 
        for i = 1:problem.nodes
            
            epsilon_span = [(i - 1), i]*(tspan); % [/]
            % Risoluzione del sistema di equazioni differenziali per l'intervallo corrente
            [epsilon, y] = ode45(@(epsilon,z) satelliteODEE(epsilon, z, controlRet(i), tf, problem, psi(i), TRet(i)), epsilon_span , y_prev);
        
            % Aggiunta dei risultati correnti ai risultati complessivi
            t_combined = [t_combined; epsilon(end)]; % adim
            y_combined = [y_combined; y(end,:)]; % adim
            % Aggiornamento delle condizioni iniziali per la prossima iterazione
            y_prev = y(end, :)';
        end
    y_combined(end,7)
    t_final = t_combined*t_rif;%[s]
    y_final = y_combined; %[/]
    
    tspan=(tf1-tf)/problem.nodes1;
    Tx=[];
    Tz=[];
    for i = problem.nodes+1:problem.nodes+problem.nodes1
            
            epsilon_span = [(i - 1)-problem.nodes+tf, i-problem.nodes+tf]*(tspan); % [/]
            % Risoluzione del sistema di equazioni differenziali per l'intervallo corrente
            [epsilon, y] = ode45(@(epsilon,z) satelliteODEE(epsilon, z, controlRet(i), tf, problem, psi(i), 4000/(m_rif*a_rif)), epsilon_span , y_prev);
        
            % Aggiunta dei risultati correnti ai risultati complessivi
            t_combined = [t_combined; tf+epsilon(end)]; % adim
            y_combined = [y_combined; y(end,:)]; % adim
            % Aggiornamento delle condizioni iniziali per la prossima iterazione
            y_prev = y(end, :)';  
            %y_combined(end,7)
    end
   
    t_combined;
    t_final1 = t_combined*t_rif;%[s]
    y_final1 = y_combined; %[/]

    obj = feval(@costt, tf, y_final1(end, 7))
    
solutionOut.tf_tf1=[tf,tf1];
solutionOut.time = t_combined'; % Tempo totale
solutionOut.control = [controlRet;TRet]; % Controlli ottimizzati
solutionOut.state = y_final1(:, 1:problem.ns); % Stati
solutionOut.objective = obj; % Obiettivo finale senza integrando
solutionOut.solTime = solTime; % Tempo di soluzione totale
end      

function obj = objectivee(input, problem)
global t_rif m_rif a_rif
    phi = input(1:110);
    Tfinal = input(111:210);
    tf = input(211);
    tf1= input(212);
    t_combined = [];
    y_combined = [];
    
    % Loop per l'integrazione su ogni intervallo tspan
    tspan=tf/problem.nodes;
    y_prev=problem.x0;
    nod_tot=problem.nodes+problem.nodes1;
   % T=4000.*ones(1,problem.nodes);
    psi=deg2rad(zeros(1,nod_tot)); 
        for i = 1:problem.nodes
            
            epsilon_span = [(i - 1), i]*(tspan); % [/]
            % Risoluzione del sistema di equazioni differenziali per l'intervallo corrente
            [epsilon, y] = ode45(@(epsilon,z) satelliteODEE(epsilon, z, phi(i), tf, problem, psi(i), Tfinal(i)), epsilon_span , y_prev);
        
            % Aggiunta dei risultati correnti ai risultati complessivi
            t_combined = [t_combined; epsilon(end)]; % adim
            y_combined = [y_combined; y(end,:)]; % adim
            % Aggiornamento delle condizioni iniziali per la prossima iterazione
            y_prev = y(end, :)';
        end
    
    t_final = t_combined*t_rif;%[s]
    y_final = y_combined; %[/]
    
    tspan=(tf1-tf)/problem.nodes1;
    for i = problem.nodes+1:problem.nodes+problem.nodes1
            
            epsilon_span = [(i - 1)-problem.nodes+tf, i-problem.nodes+tf]*(tspan); % [/]
            % Risoluzione del sistema di equazioni differenziali per l'intervallo corrente
            [epsilon, y] = ode45(@(epsilon,z) satelliteODEE(epsilon, z, phi(i) , tf, problem, psi(i), 4000/(m_rif*a_rif)), epsilon_span , y_prev);
        
            % Aggiunta dei risultati correnti ai risultati complessivi
            t_combined = [t_combined; tf+epsilon(end)]; % adim
            y_combined = [y_combined; y(end,:)]; % adim
            % Aggiornamento delle condizioni iniziali per la prossima iterazione
            y_prev = y(end, :)';
    end
    t_combined;
    t_final1 = t_combined*t_rif;%[s]
    y_final1 = y_combined; %[/]

    obj = feval(@costt, tf, y_final1(end, 7));
    
end

function obj = costt(tf, xf)
    obj = -xf;
end


function dydt = satelliteODEE(epsilon, y, phi, tf, problem, psi, T) % satelliteODE(epsilon [/], z[/], T(i) [N], phi(i) [/], psi(i)[/], tspan[s])

    g_moon=1.623; %m/s^2
    c=290*9.80665;       %m/s
    % Adimensionalizzazione
    r_rif = 15000; %m
    v_rif = 1000;  %m/s
    m_rif = 1600;  %kg
    a_rif = v_rif^2./ r_rif; %m/s^2  =186
    g_moon_adim = g_moon / a_rif; %[/] =0.0087
    c_adim=c/v_rif;
    t_rif=(r_rif/v_rif); %s
     %N= 0.0179
  
    dydt = zeros(7, 1);
    dydt(1) = y(4);                          % Equazione per dx1/dt; Vx_adim
    dydt(2) = y(5);                          % Equazione per dx2/dt; Vy_adim
    dydt(3) = y(6);
    dydt(4) = (T/y(7))*cos(phi)*cos(psi);    % a_x_adim 
    dydt(5) = (T/y(7))*cos(phi)*sin(psi);    % a_y_adim
    dydt(6) = (T/y(7))*sin(phi)-g_moon_adim; % a_z_adim
    dydt(7) = -T/c_adim;                     

end


function [c, ceq] = constraintss(input, problem)
    global t_rif r_rif m_rif a_rif
    phi = input(1:110);
    Tfinal = input(111:210);
    tf = input(211);
    tf1= input(212);
    t_combined = [];
    y_combined = [];
    
    %vincolo di disuguaglianza. ottimizzazione richiede c<=0; c=tf−tf1 fa sì che tf1>tf
    c(1)=tf-tf1;
    
    ceq(1) = tf1-tf-5./t_rif;

    % Loop per l'integrazione su ogni intervallo tspan
    tspan=tf/problem.nodes;
    y_prev=problem.x0;
    nod_tot=problem.nodes+problem.nodes1;
    %T=4000.*ones(1,nod_tot);
    psi=deg2rad(zeros(1,nod_tot)); 
        for i = 1:problem.nodes
            
            epsilon_span = [(i - 1), i]*(tspan); % [/]
            % Risoluzione del sistema di equazioni differenziali per l'intervallo corrente
            [epsilon, y] = ode45(@(epsilon,z) satelliteODEE(epsilon, z, phi(i), tf, problem, psi(i), Tfinal(i)), epsilon_span , y_prev);
        
            % Aggiunta dei risultati correnti ai risultati complessivi
            t_combined = [t_combined; epsilon(end)]; % adim
            y_combined = [y_combined; y(end,:)]; % adim
            % Aggiornamento delle condizioni iniziali per la prossima iterazione
            y_prev = y(end, :)';
        end
    
    t_final = t_combined*t_rif;%[s]
    y_final = y_combined; %[/]
    
    ceq(2) = y_final(end, 4); % Vx = 0 a t_att = 5s
   
    tspan=(tf1-tf)/problem.nodes1;
    for i = problem.nodes+1:problem.nodes+problem.nodes1
            epsilon_span = [(i - 1)-problem.nodes+tf, i-problem.nodes+tf]*(tspan); % [/]
            % Risoluzione del sistema di equazioni differenziali per l'intervallo corrente
            [epsilon, y] = ode45(@(epsilon,z) satelliteODEE(epsilon, z, phi(i), tf, problem, psi(i), 4000/(m_rif*a_rif)), epsilon_span , y_prev);
        
            % Aggiunta dei risultati correnti ai risultati complessivi
            t_combined = [t_combined; tf+epsilon(end)]; % adim
            y_combined = [y_combined; y(end,:)]; % adim
            % Aggiornamento delle condizioni iniziali per la prossima iterazione
            y_prev = y(end, :)';

    end
    t_final1 = t_combined*t_rif;%[s]
    y_final1 = y_combined; %[/]

    ceq(3) = y_final1(end, 3); % z = 0
    ceq(4) = y_final1(end, 6); % Vz = 0
end